@@ -23,15 +23,11 @@
 Schema dictionaries
 """
 
+from collections import OrderedDict
 from inspect import isclass
 from numbers import Number
 
 
-class Nothing:
-    """Special 'type' used for checks"""
-    pass
-
-
 class SchemaError(Exception):
     """Raised if the schema dictionary is ill-defined"""
     pass
@@ -39,66 +35,110 @@ class SchemaError(Exception):
 
 class Validators:
     """
-    Collection of validators
+    Collection of validator functions
+
+    All validator functions must accept three arguments in the order listed
+    below. The actual variable names may differ depending on context.
+
+    Args:
+        :value: value to be checked
+        :exp_value: expected value or comparison object
+        :key: dictionary key (used in error message)
     """
 
-    @staticmethod
-    def is_type(value, exp_type, err_msg=''):
-        if isinstance(exp_type, dict):
-            pass
-            # TODO
-        if not isinstance(value, exp_type):
-            raise TypeError(err_msg)
+    @classmethod
+    def is_type(cls, value, exp_type, key):
+        # TODO????
+        # if isinstance(exp_type, dict):
+        #     cls.check_schemadict(value, exp_type, key)
+
+        # Note: isinstance(True, int) evaluates to True
+        if type(value) not in (exp_type,):
+            raise TypeError(
+                f"unexpected type for {key!r}: " +
+                f"expected {exp_type!r}, but was {type(value)}"
+            )
 
     @staticmethod
-    def is_gt(value, comp_value, err_msg=''):
+    def is_gt(value, comp_value, key):
         if not value > comp_value:
-            raise ValueError(err_msg)
+            raise ValueError(
+                f"{key!r} too small: " +
+                f"expected > {comp_value!r}, but was {value!r}"
+            )
 
     @staticmethod
-    def is_lt(value, comp_value, err_msg=''):
+    def is_lt(value, comp_value, key):
         if not value < comp_value:
-            raise ValueError(err_msg)
+            raise ValueError(
+                f"{key!r} too large: " +
+                f"expected < {comp_value!r}, but was {value!r}"
+            )
 
-    @staticmethod
-    def is_le(value, comp_value, err_msg=''):
-        if not value <= comp_value:
-            raise ValueError(err_msg)
-
-    @staticmethod
-    def is_ge(value, comp_value, err_msg=''):
+    @classmethod
+    def is_ge(cls, value, comp_value, key):
         if not value >= comp_value:
-            raise ValueError(err_msg)
+            raise ValueError(
+                f"{key!r} too small: " +
+                f"expected >= {comp_value!r}, but was {value!r}"
+            )
 
     @staticmethod
-    def has_min_len(value, min_len, err_msg=''):
+    def is_le(value, comp_value, key):
+        if not value <= comp_value:
+            raise ValueError(
+                f"{key!r} too large: " +
+                f"expected <= {comp_value!r}, but was {value!r}"
+            )
+
+    @staticmethod
+    def has_min_len(value, min_len, key):
         if not len(value) >= min_len:
-            raise ValueError(err_msg)
+            raise ValueError(
+                f"length of {key!r} too small: " +
+                f"expected >= {min_len!r}, but was {len(value)!r}"
+            )
 
     @staticmethod
-    def has_max_len(value, max_len, err_msg=''):
+    def has_max_len(value, max_len, key):
         if not len(value) <= max_len:
-            raise ValueError(err_msg)
+            raise ValueError(
+                f"length of {key!r} too large: " +
+                f"expected <= {max_len!r}, but was {len(value)!r}"
+            )
 
     @staticmethod
-    def check_item_types(iterable, exp_item_type, err_msg=''):
+    def check_item_types(iterable, exp_item_type, key):
         if not all(isinstance(item, exp_item_type) for item in iterable):
-            raise TypeError(err_msg)
+            raise TypeError(
+                f"unexpected type for item in iterable {key!r}: " +
+                f"expected {exp_item_type!r}"
+            )
+
+    @classmethod
+    def check_item_schema(cls, iterable, item_schema, key):
+        for item in iterable:
+            cls.check_schemadict(item, item_schema, key)
+
+    @staticmethod
+    def check_schemadict(testdict, schema, key):
+        schemadict(schema).validate(testdict)
 
 
-class ValidatorDict(dict):
+class ValidatorDict(OrderedDict):
     """
-    Map 'type' (=key) and validator functions
+    Use to map 'type' (=key) and validator functions
 
-    Raise 'SchemaError' if meta schema for 'type' is not defined
+    Raise 'SchemaError' if meta schema for 'type' is not defined.
     """
-
     def __missing__(self, key):
-        raise SchemaError(f"Meta schema not defined for {key!r}")
+        raise SchemaError(f"meta-schema not defined for {key!r}")
 
 
+# Check type (required by all validators)
 _VAL_TYPE = {'type': Validators.is_type}
 
+# Check numerical relations (int, float, Number)
 _VAL_NUM_REL = {
     **_VAL_TYPE,
     '>': Validators.is_gt,
@@ -107,16 +147,26 @@ _VAL_NUM_REL = {
     '<=': Validators.is_le,
 }
 
+# Check countable objects (list, tuple, str)
 _VAL_COUNTABLE = {
     **_VAL_TYPE,
     'min_len': Validators.has_min_len,
     'max_len': Validators.has_max_len,
 }
+
+# Check iterable objects (list, tuple)
 _VAL_ITERABLE = {
     **_VAL_COUNTABLE,
     'item_types': Validators.check_item_types,
+    'item_schema': Validators.check_item_schema,
 }
 
+_VAL_SUBSCHEMA = {
+    **_VAL_TYPE,
+    'schema': Validators.check_schemadict,
+}
+
+# Validators for primitive types
 STANDARD_VALIDATORS = ValidatorDict({
     bool: _VAL_TYPE,
     int: _VAL_NUM_REL,
@@ -124,79 +174,89 @@ STANDARD_VALIDATORS = ValidatorDict({
     str: _VAL_COUNTABLE,
     list: _VAL_ITERABLE,
     tuple: _VAL_ITERABLE,
+    dict: _VAL_SUBSCHEMA,
+    # schemadict: _VAL_SUBSCHEMA,
 })
 
 
 class schemadict(dict):
     """
-    A 'schemadict' is a dictionary that defines an expected format for how some
-    other dictionary should be formatted. Schemadicts provide a 'validate()'
-    method that checks if the test dictionary is conform with the schema.
-
-    The schema validation is loosely inspired by JSON schema, see:
-
-    * https://json-schema.org/understanding-json-schema/reference/index.html
+    A *schemadict* is a regular Python dictionary which specifies the type and
+    format of values for some given key. To check if a test dictionary is
+    conform with the expected schema, *schemadict* provides the `validate()`
+    method. If the test dictionary is ill-defined, an error will be thrown,
+    otherwise `None` is returned.
     """
 
+    # Special 'type' used for checks
+    class Nothing:
+        pass
+
+    # ===== TODO (START) =====
+    # TODO: find better solution!?
     SPECIAL_KEY_CHECK_REQ_KEYS = '__REQUIRED_KEYS'
-    SPECIAL_KEY_LITERAL_EQUAL = '__EQUAL'
+    # SPECIAL_KEY_LITERAL_EQUAL = '__EQUAL'
 
     SPECIAL_KEYS = [
         SPECIAL_KEY_CHECK_REQ_KEYS,
-        SPECIAL_KEY_LITERAL_EQUAL,
+        # SPECIAL_KEY_LITERAL_EQUAL,
     ]
 
     SPECIAL_VALUE_ANY_CLASS_TYPE = '__ANY_CLASS_TYPE'
     SPECIAL_VALUES = [
         SPECIAL_VALUE_ANY_CLASS_TYPE,
     ]
+    # ===== TODO (END) =====
+
+    # ===== TODO (START) =====
+    # TODO: add type dict and 'schema' key
+    # TODO: write more succinct
+    # TODO: add 'default' key (can be None, same type as 'type', callable...)
 
     # Define the allowed schema in terms of a schema dict
-    # TODO: rethink --> validator objects instead !?
-    # TODO: add type dict and 'schema' key
-    # TODO: add 'default' key (can be None, same type as 'type', callable...)
-    META_SCHEMA = {
-        bool: {
-            SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
-            'type': {SPECIAL_KEY_LITERAL_EQUAL: bool},
-        },
-        int: {
-            SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
-            'type': {SPECIAL_KEY_LITERAL_EQUAL: int},
-            '>': {'type': Number},
-            '<': {'type': Number},
-            '<=': {'type': Number},
-            '>=': {'type': Number},
-        },
-        float: {
-            SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
-            'type': {SPECIAL_KEY_LITERAL_EQUAL: float},
-            '>': {'type': Number},
-            '<': {'type': Number},
-            '<=': {'type': Number},
-            '>=': {'type': Number},
-        },
-        str: {
-            SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
-            'type': {SPECIAL_KEY_LITERAL_EQUAL: str},
-            'min_len': {'type': Number},
-            'max_len': {'type': Number},
-        },
-        list: {
-            SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
-            'type': {SPECIAL_KEY_LITERAL_EQUAL: list},
-            'min_len': {'type': Number},
-            'max_len': {'type': Number},
-            'item_types': {'type': SPECIAL_VALUE_ANY_CLASS_TYPE},
-        },
-        tuple: {
-            SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
-            'type': {SPECIAL_KEY_LITERAL_EQUAL: tuple},
-            'min_len': {'type': Number},
-            'max_len': {'type': Number},
-            'item_types': {'type': SPECIAL_VALUE_ANY_CLASS_TYPE},
-        },
-    }
+    # META_SCHEMA = {
+    #     bool: {
+    #         SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
+    #         'type': {SPECIAL_KEY_LITERAL_EQUAL: bool},
+    #     },
+    #     int: {
+    #         SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
+    #         'type': {SPECIAL_KEY_LITERAL_EQUAL: int},
+    #         '>': {'type': Number},
+    #         '<': {'type': Number},
+    #         '<=': {'type': Number},
+    #         '>=': {'type': Number},
+    #     },
+    #     float: {
+    #         SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
+    #         'type': {SPECIAL_KEY_LITERAL_EQUAL: float},
+    #         '>': {'type': Number},
+    #         '<': {'type': Number},
+    #         '<=': {'type': Number},
+    #         '>=': {'type': Number},
+    #     },
+    #     str: {
+    #         SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
+    #         'type': {SPECIAL_KEY_LITERAL_EQUAL: str},
+    #         'min_len': {'type': Number},
+    #         'max_len': {'type': Number},
+    #     },
+    #     list: {
+    #         SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
+    #         'type': {SPECIAL_KEY_LITERAL_EQUAL: list},
+    #         'min_len': {'type': Number},
+    #         'max_len': {'type': Number},
+    #         'item_types': {'type': SPECIAL_VALUE_ANY_CLASS_TYPE},
+    #     },
+    #     tuple: {
+    #         SPECIAL_KEY_CHECK_REQ_KEYS: ['type'],
+    #         'type': {SPECIAL_KEY_LITERAL_EQUAL: tuple},
+    #         'min_len': {'type': Number},
+    #         'max_len': {'type': Number},
+    #         'item_types': {'type': SPECIAL_VALUE_ANY_CLASS_TYPE},
+    #     },
+    # }
+    # ===== TODO (END) =====
 
     VALIDATORS = STANDARD_VALIDATORS
 
@@ -206,13 +266,14 @@ class schemadict(dict):
 
         elif key == self.SPECIAL_KEY_LITERAL_EQUAL:
             if value != testdict[key]:
+                # ===== TODO =====
                 # TODO: Add error message
                 raise ValueError
 
-    def _check_special_values(self, key, value, testdict):
-        if value == self.SPECIAL_VALUE_ANY_CLASS_TYPE:
-            if not isclass(testdict[key]):
-                raise ValueError(f"Value for key {key!r} must be a class")
+    # def _check_special_values(self, key, value, testdict):
+    #     if value == self.SPECIAL_VALUE_ANY_CLASS_TYPE:
+    #         if not isclass(testdict[key]):
+    #             raise ValueError(f"Value for key {key!r} must be a class")
 
     @staticmethod
     def check_req_keys_in_dict(req_keys, testdict):
@@ -227,6 +288,7 @@ class schemadict(dict):
             :KeyError: If a required key is not found in the test dictionary
         """
 
+        # ===== TODO =====
         # TODO: check that req_keys is list of strings !?
 
         testdict_keys = list(testdict.keys())
@@ -257,51 +319,36 @@ class schemadict(dict):
     def _validate(self, testdict):
         """Validate 'testdict' against 'self' (without schema validation)"""
 
-        for schemadict_key, schemadict_value in self.items():
-            # Treat special keys separately
-            if schemadict_key in self.SPECIAL_KEYS:
-                self._check_special_key(schemadict_key, schemadict_value, testdict)
+        for sd_key, sd_value in self.items():
+            # ===== TODO (START) =====
+            # TODO: find better solution
+            # Treat special keys/values separately
+            if sd_key in self.SPECIAL_KEYS:
+                self._check_special_key(sd_key, sd_value, testdict)
                 continue
+###            if sd_value in self.SPECIAL_VALUES:
+###                self._check_special_values(sd_key, sd_value, testdict)
+###                continue
+            # ===== TODO (END) =====
 
-            # Treat special values separately
-            if schemadict_value in self.SPECIAL_VALUES:
-                self._check_special_values(schemadict_key, schemadict_value, testdict)
-                continue
-
-            testdict_value = testdict.get(schemadict_key, None)
-
-            # Continue if testdict does not have corresponding schemadict_value.
+            td_value = testdict.get(sd_key, None)
+            # Continue if testdict does not have corresponding sd_value.
             # Note that required keys are checked separately.
-            if testdict_value is None:
+            if td_value is None:
                 continue
 
-            # Basic type check (every schema entry must define a type)
-            schemadict_type = schemadict_value.get('type', None)
-            # ===== 'type' =====
-            Validators.is_type(
-                testdict_value,
-                schemadict_type,
-                f"Unexpected type for key {schemadict_key!r}. " +
-                f"Expected {schemadict_type}, got {type(testdict_value)}."
-            )
-
-            # TYPE: dict
-            if schemadict_type is dict:
-                sub_schemadict = schemadict_value.get('schema', None)
-                if sub_schemadict is not None:
-                    schemadict(sub_schemadict).validate(testdict_value)
-                    continue
-
-            for validator_key, validator in self.VALIDATORS[schemadict_type].items():
-                exp_value = schemadict_value.get(validator_key, None)
+            for validator_key, validator in self.VALIDATORS[sd_value['type']].items():
+                exp_value = sd_value.get(validator_key, None)
                 if exp_value is not None:
-                    validator(testdict_value, exp_value, err_msg='')
+                    validator(td_value, exp_value, sd_key)
 
+    # ===== TODO =====
+    # TODO: continue here --------------------------------------------------
     def get_default_value_dict(self):
         """
         Return a dictionary with default values based on a schema dict
 
-        Default values are genereated as follows:
+        Default values are generated as follows:
             * If a key 'default' exists, the corresponding value will be used
                 * If value is callable object, then the called value will be used
                 * If value is non-callable, the value itself will be used
@@ -311,6 +358,7 @@ class schemadict(dict):
         Example:
 
         .. code:: python
+
             from datetime import datetime
 
             def time_now():
@@ -340,6 +388,7 @@ class schemadict(dict):
         The function will return
 
         .. code:: python
+
             defaults = {
                 'time': '08:40',
                 'person': 'C.Lindbergh',
@@ -368,8 +417,8 @@ class schemadict(dict):
 
             # ----- Set default value -----
             # Hint: default value could be intentionally set to None
-            default_value = value.get('default', Nothing())
-            if not isinstance(default_value, Nothing):
+            default_value = value.get('default', self.Nothing())
+            if not isinstance(default_value, self.Nothing):
                 if callable(default_value):
                     defaults[key] = default_value()
                 else:
